from django.conf import settings
from django.http.response import JsonResponse
from django.utils import translation

from babel import Locale

from pfx.pfxcore.decorator import rest_api, rest_view

from .rest_views import BaseRestView


def get_locales():
    return [
        dict(
            pk=code,
            name=Locale.parse(translation.to_locale(code)).get_display_name())
        for code, __ in settings.LANGUAGES]


@rest_view("/locales")
class LocaleRestView(BaseRestView):
    @rest_api('/languages', public=True)
    def locales(self):
        locales = get_locales()
        return JsonResponse({
            'items': locales, 'meta': {'count': len(locales)}})



from unittest.mock import patch

from django.test import TestCase
from django.test.utils import override_settings

from pfx.pfxcore.test import APIClient, TestAssertMixin


class BasicAPIErrorTest(TestAssertMixin, TestCase):

    def setUp(self):
        self.client = APIClient(default_locale='en')

    @classmethod
    def setUpTestData(cls):
        pass

    @override_settings(DEBUG=False)
    def test_resource_not_found(self):
        response = self.client.get('/api/error/404')
        self.assertRC(response, 404)

    @override_settings(DEBUG=False)
    @patch('builtins.print')
    def test_error500(self, mock_print):
        response = self.client.get('/api/error/500')
        self.assertRC(response, 500)



import logging

from django.conf import settings
from django.http import HttpResponseRedirect
from django.utils import translation
from django.utils.cache import patch_vary_headers
from django.utils.deprecation import MiddlewareMixin

logger = logging.getLogger(__name__)


def get_language_from_request(request):
    """Custom version of django get_language_from_request from translation.

    Remove path and cookies parsing and just use X-Custom-language as
    first choice if it is defined and valid, then use Accept-Language.
    """
    _trans = translation._trans  # Load _trans dynamically from translation.

    custom_lang = translation.to_language(
        request.META.get('HTTP_X_CUSTOM_LANGUAGE', ''))
    try:
        return _trans.get_supported_language_variant(custom_lang)
    except LookupError:
        pass

    accept = request.META.get('HTTP_ACCEPT_LANGUAGE', '')
    for accept_lang, unused in _trans.parse_accept_lang_header(accept):
        if accept_lang == '*':
            break  # pragma: no cover
        if not _trans.language_code_re.search(accept_lang):
            continue  # pragma: no cover
        try:
            return _trans.get_supported_language_variant(accept_lang)
        except LookupError:
            continue

    try:
        return _trans.get_supported_language_variant(settings.LANGUAGE_CODE)
    except LookupError:  # pragma: no cover
        return settings.LANGUAGE_CODE


class LocaleMiddleware(MiddlewareMixin):
    response_redirect_class = HttpResponseRedirect

    def process_request(self, request):
        language = get_language_from_request(request)
        translation.activate(language)
        request.LANGUAGE_CODE = translation.get_language()

    def process_response(self, request, response):
        patch_vary_headers(response, ('Accept-Language',))
        code = (
            len(request.LANGUAGE_CODE) == 5 and
            request.LANGUAGE_CODE[:3] + request.LANGUAGE_CODE[-2:].upper() or
            request.LANGUAGE_CODE)
        response.headers.setdefault('Content-Language', code)
        return response



from .authentication_views import (
    AuthenticationView,
    ForgotenPasswordView,
    SignupView,
)
from .filters_views import Filter, FilterGroup, ModelFilter
from .locale_views import LocaleRestView
from .rest_views import (
    BaseRestView,
    BodyMixin,
    CreateRestViewMixin,
    DeleteRestViewMixin,
    DetailRestViewMixin,
    ListRestViewMixin,
    ModelBodyMixin,
    ModelMixin,
    ModelResponseMixin,
    RestView,
    SecuredRestViewMixin,
    UpdateRestViewMixin,
    resource_not_found,
)



import json
import logging
from datetime import date

from django.db import IntegrityError, transaction
from django.db.models import DateField, ForeignKey, Manager, Model, Q
from django.http import JsonResponse
from django.utils.formats import date_format
from django.utils.translation import gettext_lazy as _
from django.views import View

from pfx.pfxcore.decorator import rest_api
from pfx.pfxcore.exceptions import APIError, ForbiddenError, UnauthorizedError
from pfx.pfxcore.model_resolver import (
    MetaResolver,
    ObjectResolver,
    ResolverValidationError,
)
from pfx.pfxcore.models import JSONReprMixin
from pfx.pfxcore.shortcuts import f, get_object, parse_bool

logger = logging.getLogger(__name__)

LIST_META = ['count', 'pagination']


# HTTP 404 handler
def resource_not_found(request, exception):
    return JsonResponse(dict(message=_(
        "Resource not found")), status=404)


class ModelMixin:
    queryset = None
    fields = []

    def get_object(self, **kwargs):
        return get_object(self.get_queryset(), **kwargs)

    def get_queryset(self):
        return self.queryset

    def get_list_queryset(self):
        return self.get_queryset()

    def get_fields(self):
        return self.fields or [f.name for f in self.model._meta.fields]

    def get_model_fields(self):
        for lookup in self.get_fields():
            yield lookup, MetaResolver(self.model).get_field(lookup)

    @property
    def model(self):
        return self.get_queryset().model

    @property
    def model_name(self):
        return self.model._meta.verbose_name

    @property
    def date_format(self):
        return parse_bool(self.request.GET.get('date_format'))

    def _json_object(self, obj):
        if isinstance(obj, JSONReprMixin):
            return obj.json_repr()
        return dict(pk=obj.pk, resource_name=str(obj))

    def message_response(self, message, **kwargs):
        return JsonResponse(dict(message=message, **kwargs))


class ModelResponseMixin(ModelMixin):
    def serialize_field(self, obj, field_name):
        value = ObjectResolver(obj).get_value(field_name)
        field = MetaResolver(self.model).get_field(field_name)
        if self.date_format and isinstance(value, date):
            return dict(
                value=value,
                formatted=date_format(
                    value, format='SHORT_DATE_FORMAT', use_l10n=True))
        if isinstance(value, Model):
            return self._json_object(value)
        if isinstance(value, Manager):
            return [self._json_object(o) for o in value.all()]
        if hasattr(field, 'choices') and field.choices:
            choices = {k: v for k, v in field.choices}
            if value in choices:
                return dict(value=value, label=_(choices[value]))
            else:  # pragma: no cover
                return None
        return value

    def serialize_object(self, obj, **fields):
        vals = self._json_object(obj)
        vals.update(fields)
        return vals

    def response(self, o, **meta):
        return JsonResponse(self.serialize_object(o, **{
            f: self.serialize_field(o, f)
            for f in self.get_fields()}, meta=meta))

    def validate(self, resolver, created=False, **kwargs):
        resolver.validate(**kwargs)

    def is_valid(self, resolver, created=False):
        resolver.save()
        message = (
                created and
                _("{model} {obj} created.") or
                _("{model} {obj} updated."))
        return self.response(
            resolver.object, message=f(
                message, model=self.model_name, obj=resolver.object))

    def is_invalid(self, resolver, errors):
        return JsonResponse(errors, status=422)

    def field_meta(self, field):
        name = field.name
        if hasattr(field, 'verbose_name'):
            name = field.verbose_name
        elif hasattr(field, 'related_model'):
            if (hasattr(field, 'multiple') and field.multiple and
                    hasattr(field.related_model._meta, 'verbose_name_plural')):
                name = field.related_model._meta.verbose_name_plural
            elif hasattr(
                field.related_model._meta, 'verbose_name'
            ):  # pragma: no cover
                name = field.related_model._meta.verbose_name
        res = {'type': field.get_internal_type(), 'name': name}
        if hasattr(field, 'choices') and field.choices:
            res['choices'] = [
                dict(label=_(v), value=k) for k, v in field.choices]
        return res

    def object_meta(self):
        return {n: self.field_meta(f) for n, f in self.get_model_fields()}

    @rest_api("/meta", method="get")
    def get_meta(self, *args, **kwargs):
        return JsonResponse(self.object_meta())


class BodyMixin:
    def deserialize_body(self):
        return json.loads(self.request.body)


class ModelBodyMixin(BodyMixin, ModelMixin):
    readonly_fields = []

    def get_readonly_fields(self, obj=None):
        return self.readonly_fields

    def _field_data(self, field_name, value):
        field = MetaResolver(self.model).get_field(field_name)
        if isinstance(field, ForeignKey):
            pk = (value['pk'] if isinstance(value, dict) and 'pk' in value
                  else value)
            return f'{field_name}_id', pk or None
        if hasattr(field, 'choices') and field.choices:
            value = (value['value']
                     if isinstance(value, dict) and 'value' in value
                     else value)
        elif isinstance(field, DateField):
            value = (value['value']
                     if isinstance(value, dict) and 'value' in value
                     else value)
        return field_name, value

    def get_model_data(self, obj, data, created):
        data.pop('id', None)
        data.pop('pk', None)
        ro_fields = self.get_readonly_fields(obj)
        return dict(
            self._field_data(k, v) for k, v in data.items()
            # Todo: throw an error for non existing fields.
            if k in self.get_fields() and
            # Todo: Log a warning for readonly ignores fields.
            k not in ro_fields and
            # Deny API changes on technical django _id and _pk fields.
            k[-3:] not in ('_id', '_pk'))


class ListRestViewMixin(ModelResponseMixin):
    list_fields = []
    filters = []

    def get_list_fields(self):
        return self.list_fields or self.get_fields()

    def parse_list_meta(self):
        meta_arg = self.request.GET.get('meta', 'all')
        meta = meta_arg.split(',') or []
        if 'all' in meta:
            return LIST_META
        return meta

    def _list_meta_count(self):
        return self.get_list_queryset().count()

    def _list_meta_pagination(self):
        return dict(
            page_size=int(self.request.GET.get('page_size', 10)),
            page=int(self.request.GET.get('page', 1)),
            page_subset=int(self.request.GET.get('page_subset', 5)))

    def _list_meta_subset(self):
        return dict(
            limit=int(self.request.GET.get('limit', 10)),
            offset=int(self.request.GET.get('offset', 0)))

    def build_list_meta(self):
        def _meta(meta):
            fn = getattr(self, f'_list_meta_{meta}', None)
            if not callable(fn):
                raise APIError(
                    _("Meta {meta} does not exists.").format(meta=meta))
            return fn

        return {
            meta: _meta(meta)()
            for meta in self.parse_list_meta()}

    def search_filter(self, search):  # pragma: no cover
        return Q()

    @rest_api("/filters", method="get")
    def get_filter(self, *args, **kwargs):
        return JsonResponse({
            'items': [f.meta for f in self.filters]})

    def get_list_queryset(self):
        qs = super().get_list_queryset()
        search = self.request.GET.get('search')
        for filter in self.filters:
            qs = qs.filter(filter.query(self.request.GET))
        if search:
            qs = qs.filter(
                self.search_filter(search))
        order = self.request.GET.get('order')
        if order:
            qs = qs.order_by(*order.split(','))
        return qs

    def get_list_result(self, qs):
        for o in qs:
            yield self.serialize_object(o, **{
                f: self.serialize_field(o, f)
                for f in self.get_list_fields()})

    @rest_api("", method="get")
    def get_list(self, *args, **kwargs):
        res = {}
        meta = self.build_list_meta()
        qs = self.get_list_queryset()
        if 'pagination' in meta:
            count = qs.count()
            pagination = meta['pagination']
            page = pagination['page']
            limit = pagination['page_size']
            page_count = (1 + (count - 1) // limit) or 1
            offset = (page - 1) * limit
            subset = pagination['page_subset']
            subset_first = min(
                max(page - subset // 2, 1), max(page_count - subset + 1, 1))
            qs = qs.all()[offset:offset + limit]
            pagination.update(dict(
                count=count,
                page_count=page_count,
                subset=list(range(
                    subset_first,
                    min(subset_first + subset, page_count + 1)))))
        elif 'subset' in meta:
            count = qs.count()
            subset = meta['subset']
            limit = subset['limit']
            page_count = (1 + (count - 1) // limit) or 1
            offset = subset['offset']
            qs = qs.all()[offset:offset + limit]
            subset.update(dict(
                count=count,
                page_count=page_count,
                limit=limit,
                offset=offset))
        else:
            qs = qs.all()
        if meta:
            res['meta'] = meta
        res['items'] = list(self.get_list_result(qs))
        return JsonResponse(res)


class DetailRestViewMixin(ModelResponseMixin):

    @rest_api("/<int:id>", method="get")
    def get(self, id, *args, **kwargs):
        obj = self.get_object(pk=id)
        return self.response(obj)


class SlugDetailRestViewMixin(ModelResponseMixin):
    SLUG_FIELD = "slug"

    @rest_api("/slug/<slug:slug>", method="get")
    def get_by_slug(self, slug, *args, **kwargs):
        obj = self.get_object(**{self.SLUG_FIELD: slug})
        return self.response(obj)


class CreateRestViewMixin(ModelBodyMixin, ModelResponseMixin):
    default_values = {}

    def get_default_values(self):
        return dict(self.default_values)

    def new_object(self):
        return self.model(**self.get_default_values())

    @rest_api("", method="post")
    def post(self, *args, **kwargs):
        obj = self.new_object()
        data = self.get_model_data(obj, self.deserialize_body(), created=True)
        resolver = ObjectResolver(obj)
        resolver.set_values(**data)
        try:
            self.validate(resolver, created=True)
            return self.is_valid(resolver, created=True)
        except ResolverValidationError as e:
            return self.is_invalid(resolver, errors=e.errors)


class UpdateRestViewMixin(ModelBodyMixin, ModelResponseMixin):
    @rest_api("/<int:id>", method="put")
    def put(self, id, *args, **kwargs):
        obj = self.get_object(pk=id)
        data = self.get_model_data(obj, self.deserialize_body(), created=False)
        resolver = ObjectResolver(obj)
        resolver.set_values(**data)
        try:
            self.validate(resolver, created=False)
            return self.is_valid(resolver, created=False)
        except ResolverValidationError as e:
            return self.is_invalid(resolver, errors=e.errors)


class DeleteRestViewMixin(ModelMixin):
    @rest_api("/<int:id>", method="delete")
    def delete(self, id, *args, **kwargs):
        obj = self.get_object(pk=id)
        try:
            with transaction.atomic():
                obj.delete()
            return self.message_response(f(
                _("{model} {obj} deleted."), model=self.model_name, obj=obj))
        except IntegrityError as e:
            logger.debug("IntegrityError: %s", e)
            raise APIError(f(_(
                "{obj} cannot be deleted because "
                "it is referenced by other objects."), obj=obj))


class SecuredRestViewMixin(View):
    default_public = False

    def perm(self):
        return True

    def _is_public(self, public, func_name):
        param = f'{func_name}_public'
        if hasattr(self, param):
            public = getattr(self, param)
        if public is None:
            public = self.default_public
        return public

    def check_perm(self, public, func_name, *args, **kwargs):
        if self._is_public(public, func_name):
            return
        if not self.request.user.is_authenticated:
            raise UnauthorizedError()
        if not self.perm():
            raise ForbiddenError()
        fperm = f'{func_name}_perm'
        if hasattr(self, fperm) and not getattr(self, fperm)(*args, **kwargs):
            raise ForbiddenError()


class BaseRestView(SecuredRestViewMixin, View):
    pass


class RestView(
        ListRestViewMixin,
        DetailRestViewMixin,
        CreateRestViewMixin,
        UpdateRestViewMixin,
        DeleteRestViewMixin,
        BaseRestView):
    pass



from datetime import date

from django.db import connection
from django.test import TestCase

from pfx.pfxcore.test import APIClient, TestAssertMixin
from tests.models import Author, Book, BookType


class BasicAPITest(TestAssertMixin, TestCase):

    def setUp(self):
        self.client = APIClient(default_locale='en')
        with connection.cursor() as cursor:
            cursor.execute("create extension if not exists unaccent;")

    @classmethod
    def setUpTestData(cls):
        cls.author1 = Author.objects.create(
            first_name='John Ronald Reuel',
            last_name='Tolkien',
            slug='jrr-tolkien')
        cls.author1_book1 = Book.objects.create(
            author=cls.author1,
            name="The Hobbit",
            pub_date=date(1937, 1, 1)
        )
        cls.author1_book2 = Book.objects.create(
            author=cls.author1,
            name="The Fellowship of the Ring",
            pub_date=date(1954, 1, 1)
        )
        cls.author1_book3 = Book.objects.create(
            author=cls.author1,
            name="The Two Towers",
            pub_date=date(1954, 1, 1)
        )
        cls.author1_book4 = Book.objects.create(
            author=cls.author1,
            name="The Return of the King",
            pub_date=date(1955, 1, 1)
        )
        cls.author2 = Author.objects.create(
            first_name='Philip Kindred',
            last_name='Dick',
            science_fiction=True,
            slug='philip-k-dick')
        cls.author3 = Author.objects.create(
            first_name='Isaac',
            last_name='Asimov',
            science_fiction=True,
            slug='isaac-asimov')
        cls.author3_book1 = Book.objects.create(
            author=cls.author3,
            name="The Caves of Steel",
            pub_date=date(1954, 1, 1),
            pages=224,
            rating=4.6,
        )
        cls.author3_book2 = Book.objects.create(
            author=cls.author3,
            name="The Naked Sun",
            pub_date=date(1957, 1, 1),
        )
        cls.author3_book3 = Book.objects.create(
            author=cls.author3,
            name="The Robots of Dawn",
            pub_date=date(1983, 1, 1),
        )
        cls.author4 = Author.objects.create(
            first_name='Joanne',
            last_name='Rowling',
            science_fiction=False,
            gender='female',
            slug='j-k-rowling')

    def test_get_list(self):
        response = self.client.get('/api/authors')

        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 4)

    def test_get_list_order_asc(self):
        response = self.client.get('/api/authors?order=last_name')

        names = [i['last_name'] for i in response.json_content['items']]
        self.assertEqual(names, ['Asimov', 'Dick', 'Rowling', 'Tolkien'])

    def test_get_list_order_desc(self):
        response = self.client.get('/api/authors?order=-last_name')

        names = [i['last_name'] for i in response.json_content['items']]
        self.assertEqual(names, ['Tolkien', 'Rowling', 'Dick', 'Asimov'])

    def test_get_list_order_multi(self):
        response = self.client.get('/api/authors?order=gender,last_name')

        names = [i['last_name'] for i in response.json_content['items']]
        self.assertEqual(names, ['Rowling', 'Asimov', 'Dick', 'Tolkien'])

    def test_get_list_order_multi_desc(self):
        response = self.client.get('/api/authors?order=gender,-last_name')

        names = [i['last_name'] for i in response.json_content['items']]
        self.assertEqual(names, ['Rowling', 'Tolkien', 'Dick', 'Asimov'])

    def test_search_list(self):
        response = self.client.get('/api/authors?search=isaac')

        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 1)
        self.assertJE(response, 'items.@0.first_name', 'Isaac')
        self.assertJE(response, 'items.@0.last_name', 'Asimov')
        self.assertJE(response, 'items.@0.gender.value', 'male')
        self.assertJE(response, 'items.@0.gender.label', 'Male')

    def test_filter_get(self):
        response = self.client.get('/api/authors/filters')
        self.assertRC(response, 200)
        self.assertJE(response, 'items.@0.name', 'book_gender')
        self.assertJE(response, 'items.@0.items.@0.name', 'science_fiction')

        response = self.client.get('/api/books/filters')
        self.assertRC(response, 200)

    def test_filter_func_bool(self):
        response = self.client.get('/api/authors?heroic_fantasy=true')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 1)
        self.assertJE(response, 'items.@0.resource_name',
                                'John Ronald Reuel Tolkien')

        response = self.client.get('/api/authors?heroic_fantasy=false')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 3)

        response = self.client.get(
            '/api/authors?heroic_fantasy=false&heroic_fantasy=true')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 4)

    def test_filter_func_date(self):
        response = self.client.get('/api/books?pub_date_gte=1955-01-01')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 3)

    def test_filter_func_char_choices(self):
        response = self.client.get('/api/authors?last_name_choices=Tolkien')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 1)

    def test_model_filter_bool(self):
        response = self.client.get('/api/authors?science_fiction=true')

        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 2)

    def test_model_filter_char(self):
        response = self.client.get('/api/authors?first_name=Isaac')

        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 1)
        self.assertJE(response, 'items.@0.first_name', "Isaac")

    def test_model_filter_integer(self):
        response = self.client.get('/api/books?pages=224')

        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 1)
        self.assertJE(response, 'items.@0.name', "The Caves of Steel")

    def test_model_filter_float(self):
        response = self.client.get('/api/books?rating=4.6')

        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 1)
        self.assertJE(response, 'items.@0.name', "The Caves of Steel")

    def test_model_filter_date(self):
        response = self.client.get('/api/books?pub_date=1954-01-01')

        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 3)

    def test_model_filter_decimal(self):
        # TODO: Wait Decimal implementation
        pass

    def test_model_filter_foreign_key(self):
        response = self.client.get(f'/api/books?author={self.author3.pk}')

        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 3)
        for item in response.json_content['items']:
            self.assertEqual(item['author']['pk'], self.author3.pk)

    def test_model_filter_foreign_key_null(self):
        book_type = BookType.objects.create(
            name="Epic Fantasy",
            slug="epic-fantasy")
        self.author1_book1.type = book_type
        self.author1_book1.save()

        response = self.client.get(
            f'/api/books?author={self.author1.pk}&type={book_type.pk}')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 1)

        response = self.client.get(
            f'/api/books?author={self.author1.pk}&type=null')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 3)
        for item in response.json_content['items']:
            self.assertEqual(item['type'], None)
        response = self.client.get(
            f'/api/books?author={self.author1.pk}&type=0')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 3)
        for item in response.json_content['items']:
            self.assertEqual(item['type'], None)
        response = self.client.get(
            f'/api/books?author={self.author1.pk}&type=')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 3)
        for item in response.json_content['items']:
            self.assertEqual(item['type'], None)

        response = self.client.get(
            f'/api/books?author={self.author1.pk}&'
            f'type={book_type.pk}&type=null')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 4)
        for item in response.json_content['items']:
            self.assertTrue(
                item['type'] is None or item['type']['pk'] == book_type.pk)

    def test_model_filter_char_choices(self):
        response = self.client.get('/api/authors?gender=male')

        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 3)

    def test_model_filter_func_bool(self):
        response = self.client.get('/api/authors?last_name=true')

        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 4)

    def test_get_list_without_pagination(self):
        response = self.client.get('/api/authors?meta=count')

        self.assertRC(response, 200)
        self.assertJE(response, 'meta.count', 4)

    def test_get_list_with_subset(self):
        for i in range(1, 11):
            Author.objects.create(
                first_name='Vladimir',
                last_name=f'Ottor {i}',
                science_fiction=True,
                slug=f'vald-ottor-{i}')

        response = self.client.get('/api/authors?meta=subset&limit=5')
        self.assertRC(response, 200)
        self.assertJE(response, 'meta.subset.offset', 0)
        self.assertJE(response, 'meta.subset.count', 14)
        item4_pk = response.json_content['items'][3]['pk']

        response = self.client.get('/api/authors?meta=subset&offset=3&limit=5')
        self.assertRC(response, 200)
        # With offset 3 the first item must be the same as the 4th
        # in the request with offest 0.
        self.assertJE(response, 'items.@0.pk', item4_pk)
        self.assertJE(response, 'meta.subset.count', 14)

    def test_wrong_meta(self):
        response = self.client.get('/api/authors?meta=qwertz')

        self.assertRC(response, 400)

    def test_meta_service(self):
        response = self.client.get('/api/authors/meta')
        self.assertRC(response, 200)
        self.assertJE(response, 'first_name.type', 'CharField')
        self.assertJE(response, 'first_name.name', 'First Name')
        self.assertJE(response, 'last_name.type', 'CharField')
        self.assertJE(response, 'last_name.name', 'Last Name')
        self.assertJE(response, 'books.type', 'ForeignKey')
        self.assertJE(response, 'books.name', 'Books')
        self.assertJE(response, 'created_at.type', 'DateField')
        self.assertJE(response, 'created_at.name', 'Created at')
        self.assertJE(response, 'name_length.type', 'IntegerField')
        self.assertJE(response, 'name_length.name', 'Name Length')

    def test_get_detail_by_id(self):
        response = self.client.get(f'/api/authors/{self.author1.pk}')
        self.assertRC(response, 200)
        self.assertJE(response, 'first_name', self.author1.first_name)
        self.assertJE(response, 'last_name', self.author1.last_name)
        self.assertJE(response, 'name_length', 25)
        self.assertJE(response, 'gender.value', 'male')
        self.assertJE(response, 'gender.label', 'Male')

    def test_get_detail_by_slug(self):
        response = self.client.get(f'/api/authors/slug/{self.author1.slug}')
        self.assertRC(response, 200)
        self.assertJE(response, 'first_name', self.author1.first_name)
        self.assertJE(response, 'last_name', self.author1.last_name)
        self.assertJE(response, 'name_length', 25)
        self.assertJE(response, 'gender.value', 'male')
        self.assertJE(response, 'gender.label', 'Male')

    def test_get_date_format(self):
        response = self.client.get(f'/api/books/{self.author1_book1.pk}')
        self.assertRC(response, 200)
        self.assertJE(response, 'pub_date', '1937-01-01')
        response = self.client.get(
            f'/api/books/{self.author1_book1.pk}?date_format=0')
        self.assertRC(response, 200)
        self.assertJE(response, 'pub_date', '1937-01-01')
        response = self.client.get(
            f'/api/books/{self.author1_book1.pk}?date_format=1')
        self.assertRC(response, 200)
        self.assertJE(response, 'pub_date.value', '1937-01-01')
        self.assertJE(response, 'pub_date.formatted', '01/01/1937')

    def test_get_non_existant_record(self):
        response = self.client.get('/api/authors/999999')
        self.assertRC(response, 404)

    def test_get_non_existant_record_by_slug(self):
        response = self.client.get('/api/authors/slug/not-existent')
        self.assertRC(response, 404)

    def test_create(self):
        response = self.client.post(
            '/api/authors', {
                'first_name': 'Arthur Charles',
                'last_name': 'Clarke',
                'name_length': 1,
                'gender': 'male',
                'slug': 'arthur-c-clarke'})

        self.assertRC(response, 200)
        self.assertJE(response, 'first_name', 'Arthur Charles')
        self.assertJE(response, 'last_name', 'Clarke')
        self.assertJE(response, 'gender.value', 'male')
        self.assertJE(response, 'gender.label', 'Male')

    def test_create_enum(self):
        response = self.client.post(
            '/api/authors', {
                'first_name': 'Arthur Charles',
                'last_name': 'Clarke',
                'name_length': 1,
                'gender': {'value': 'male'},
                'slug': 'arthur-c-clarke'})

        self.assertRC(response, 200)
        self.assertJE(response, 'first_name', 'Arthur Charles')
        self.assertJE(response, 'last_name', 'Clarke')
        self.assertJE(response, 'gender.value', 'male')
        self.assertJE(response, 'gender.label', 'Male')

    def test_update(self):
        response = self.client.put(
            f'/api/authors/{self.author1.pk}',
            {'pk': self.author2.pk,  # pk and id must be ignored
             'created_at': '2021-01-01',  # created_at must be ignored because
                                          # it is a readonly field.
             'first_name': 'J. R. R.',
             'name_length': 1,
             'gender': 'female',
             'slug': 'j-r-r-tolkien'})  # slug must be updated})

        self.assertRC(response, 200)
        self.author1.refresh_from_db()
        self.assertEqual(self.author1.first_name, 'J. R. R.')
        self.assertEqual(self.author1.last_name, self.author1.last_name)
        self.assertEqual(self.author1.slug, 'j-r-r-tolkien')
        self.assertEqual(self.author1.gender, 'female')
        self.assertNotEqual(self.author1.created_at, '2021-01-01 11:30:00')
        self.assertEqual(self.author1.created_at, self.author1.created_at)

        response = self.client.put(
            f'/api/authors/{self.author1.pk}', {'gender': {'value': 'male'}})
        self.assertRC(response, 200)
        self.author1.refresh_from_db()
        self.assertEqual(self.author1.gender, 'male')

    def test_delete(self):
        response = self.client.delete(
            f'/api/authors/{self.author2.pk}')

        self.assertRC(response, 200)

        author = Author.objects.filter(pk=self.author2.pk)
        self.assertEqual(author.count(), 0)

    def test_delete_with_wrong_key(self):
        response = self.client.delete(
            '/api/authors/99999')
        self.assertRC(response, 404)

    def test_create_with_foreignkey(self):
        response = self.client.post(
            '/api/books', {
                'name': 'The Fellowship of the Ring',
                'author': self.author1.pk,
                'pub_date': '1954-07-29',
                'created_at': '1954-07-29',
            })

        self.assertRC(response, 200)
        self.assertJE(response, 'name', 'The Fellowship of the Ring')
        self.assertJE(response, 'author.pk', self.author1.pk)
        self.assertJE(response, 'pub_date', '1954-07-29')
        self.assertNJE(response, 'created_at', '1954-07-29')

    def test_create_with_foreignkey_resource(self):
        response = self.client.post(
            '/api/books', {
                'name': 'The Fellowship of the Ring',
                'author': {
                    'pk': self.author1.pk,
                    'resource_name': "Author One"},
                'pub_date': '1954-07-29',
                'created_at': '1954-07-29',
            })

        self.assertRC(response, 200)
        self.assertJE(response, 'author.pk', self.author1.pk)

    def test_create_with_wrong_foreignkey(self):
        response = self.client.post(
            '/api/books', {
                'name': 'The Fellowship of the Ring',
                'author': 999999,
                'pub_date': '1954-07-29',
                'created_at': '1954-07-29',
            })
        self.assertRC(response, 422)
        self.assertJE(response, 'author',
                      ['Author instance with id 999999 does not exist.'])

    def test_update_with_foreignkey(self):
        book1 = Book.objects.create(
            name='The Fellowship of the Ring',
            author=self.author1,
            pub_date='1954-07-29',
        )

        response = self.client.put(
            f'/api/books/{book1.pk}', {
                'name': 'The Two Towers',
                'pub_date': '1954-11-11',
            })

        self.assertRC(response, 200)
        self.assertJE(response, 'name', 'The Two Towers')
        self.assertJE(response, 'author.pk', self.author1.pk)
        self.assertJE(response, 'pub_date', '1954-11-11')

        response = self.client.put(
            f'/api/books/{book1.pk}', {
                'name': 'The Man in the High Castle',
                'author': self.author2.pk,
                'author_id': self.author3.pk,  # must be ignored
                'pub_date': '1962-10-01',
            })

        self.assertRC(response, 200)
        self.assertJE(response, 'name', 'The Man in the High Castle')
        self.assertJE(response, 'author.pk', self.author2.pk)
        self.assertJE(response, 'pub_date', '1962-10-01')

        response = self.client.put(
            f'/api/books/{book1.pk}', {
                'name': 'A Scanner Darkly',
                'author': {
                    'pk': self.author2.pk,
                    'resource_name': 'Philip Kindred Dick'},
                'pub_date': '1977-01-01',
            })

        self.assertRC(response, 200)
        self.assertJE(response, 'name', 'A Scanner Darkly')
        self.assertJE(response, 'author.pk', self.author2.pk)
        self.assertJE(response, 'pub_date', '1977-01-01')

    def test_update_with_wrong_key_and_foreignkey(self):
        book1 = Book.objects.create(
            name='The Fellowship of the Ring',
            author=self.author1,
            pub_date='1954-07-29')

        response = self.client.put(
            '/api/books/99999', {
                'name': 'The Two Towers',
                'pub_date': '1954-11-11'})
        self.assertRC(response, 404)

        response = self.client.put(
            f'/api/books/{book1.pk}', {
                'name': 'The Two Towers',
                'author': 999999,
                'pub_date': '1954-11-11'})
        self.assertRC(response, 422)
        self.assertJE(response, 'author',
                      ['Author instance with id 999999 does not exist.'])

    def test_delete_with_foreignkey(self):
        book = Book.objects.create(
            name='The Fellowship of the Ring',
            author=self.author1,
            pub_date='1954-07-29')

        response = self.client.delete(
            f'/api/authors/{self.author1.pk}')

        self.assertRC(response, 400)
        book.refresh_from_db()

    def test_create_validation(self):
        response = self.client.post(
            '/api/books', {
                'name': '',
                'pub_date': '1954-07-29',
                'created_at': '1954-07-29',
            })

        self.assertRC(response, 422)
        self.assertJE(response, 'name',
                      ['This field cannot be blank.'])
        self.assertJE(response, 'author',
                      ['This field cannot be null.'])

    def test_update_validation(self):
        book1 = Book.objects.create(
            name='The Fellowship of the Ring',
            author=self.author1,
            pub_date='1954-07-29',
        )

        response = self.client.put(
            f'/api/books/{book1.pk}', {
                'name': '',
                'pub_date': '1954-11-11',
            })

        self.assertRC(response, 422)
        self.assertJE(response, 'name',
                      ['This field cannot be blank.'])

    def test_custom_repr(self):
        book_type = BookType.objects.create(
            name="Epic Fantasy",
            slug="epic-fantasy",
        )

        book = Book.objects.create(
            name='The Fellowship of the Ring',
            author=self.author1,
            type=book_type,
            pub_date='1954-07-29',
        )
        response = self.client.get(f'/api/books/{book.pk}')
        self.assertRC(response, 200)
        self.assertJE(response, 'type.resource_name', book_type.name)
        self.assertJE(response, 'type.resource_slug', book_type.slug)

        response = self.client.get(f'/api/book-types/{book_type.pk}')
        self.assertRC(response, 200)
        self.assertJE(response, 'resource_name', book_type.name)
        self.assertJE(response, 'resource_slug', book_type.slug)
        self.assertJE(response, 'name', book_type.name)
        self.assertJE(response, 'slug', book_type.slug)



import logging
from datetime import datetime, timedelta

from django.conf import settings
from django.contrib.auth import authenticate, get_user_model
from django.contrib.auth.tokens import default_token_generator
from django.core.exceptions import ValidationError
from django.core.mail import EmailMultiAlternatives
from django.http import JsonResponse
from django.template import loader
from django.utils.decorators import method_decorator
from django.utils.encoding import force_bytes
from django.utils.http import urlsafe_base64_decode, urlsafe_base64_encode
from django.utils.translation import gettext_lazy as _
from django.views.decorators.cache import never_cache
from django.views.decorators.debug import sensitive_post_parameters

import jwt

from pfx.pfxcore.decorator import rest_api, rest_view
from pfx.pfxcore.exceptions import AuthenticationError, ModelNotFoundAPIError
from pfx.pfxcore.models import CacheableMixin

from .rest_views import BaseRestView, BodyMixin, CreateRestViewMixin

logger = logging.getLogger(__name__)

UserModel = get_user_model()


@rest_view("/auth")
class AuthenticationView(BodyMixin, BaseRestView):
    token_generator = default_token_generator

    def login_error_response(self):
        raise AuthenticationError()

    @method_decorator(sensitive_post_parameters())
    @method_decorator(never_cache)
    @rest_api("/login", public=True, method="post")
    def login(self, *args, **kwargs):
        data = self.deserialize_body()
        user = authenticate(self.request, username=data['username'],
                            password=data['password'])
        if isinstance(user, CacheableMixin):
            user.cache_delete()   # pragma: no cover
        if user is not None:
            token = self._prepare_token(user)
            mode = self.request.GET.get('mode', 'jwt')
            if mode == 'cookie':
                if settings.PFX_TOKEN_VALIDITY:
                    expires = datetime.utcnow() + timedelta(
                        **settings.PFX_TOKEN_VALIDITY)
                else:
                    expires = None

                res = JsonResponse({
                    'user': self.get_user_information(user)
                })
                res.set_cookie('token', token, secure=True,
                               expires=expires,
                               httponly=True, samesite='None')
                return res
            else:
                return JsonResponse({
                    'token': token,
                    'user': self.get_user_information(user)
                })
        return self.login_error_response()

    @method_decorator(sensitive_post_parameters())
    @method_decorator(never_cache)
    @rest_api("/change-password", method="post")
    def change_password(self, *args, **kwargs):
        data = self.deserialize_body()
        user = authenticate(self.request,
                            username=self.request.user.get_username(),
                            password=data['old_password'])
        if user is not None:
            user.set_password(data['new_password'])
            user.save()
            return JsonResponse({
                'message': _('password updated successfully')
            })
        return JsonResponse({
            'message': 'invalid password reset'
        }, status=401)

    def _prepare_token(self, user):
        payload = {
            'pfx_user_pk': user.pk,
        }
        if settings.PFX_TOKEN_VALIDITY:
            exp = datetime.utcnow() + timedelta(**settings.PFX_TOKEN_VALIDITY)
            payload.update(exp=exp)
        payload.update(self.get_extra_payload(user))
        return jwt.encode(
            payload, settings.PFX_SECRET_KEY, algorithm="HS256")

    def get_extra_payload(self, user):
        return {}

    def get_user_information(self, user):
        return {
            'username': user.get_username(),
            'first_name': user.first_name,
            'last_name': user.last_name,
            'email': user.email}

    @method_decorator(sensitive_post_parameters())
    @method_decorator(never_cache)
    @rest_api("/set-password", public=True, method="post")
    def set_password(self, *args, **kwargs):
        data = self.deserialize_body()
        assert 'uidb64' in data and 'token' in data

        user = self.get_user(data['uidb64'])

        if (user is not None and
                self.token_generator.check_token(user, data['token'])):
            user.set_password(data['password'])
            user.is_active = True
            user.save()
            return JsonResponse({
                'message': _('password updated successfully')
            })
        raise AuthenticationError()

    def get_user(self, uidb64):
        try:
            # urlsafe_base64_decode() decodes to bytestring
            uid = urlsafe_base64_decode(uidb64).decode()
            user = UserModel._default_manager.get(pk=uid)
        except (TypeError, ValueError, OverflowError, UserModel.DoesNotExist,
                ValidationError):
            user = None
        return user


class SendMessageTokenMixin:
    email_template_name = None
    subject_template_name = None
    token_generator = default_token_generator
    extra_email_context = None
    from_email = None
    html_email_template_name = None

    def send_token_message(self, user):
        token = self.token_generator.make_token(user)
        uid = urlsafe_base64_encode(force_bytes(user.pk))
        data = {
            'target_user': user,
            'token': token,
            'uid': uid,
            'reset_url': settings.PFX_RESET_PASSWORD_URL.format(
                token=token,
                uid=uid,
            ),
            'site_name': settings.PFX_SITE_NAME,
            'user': user,
            **(self.extra_email_context or {})
        }
        subject = loader.render_to_string(self.subject_template_name, data)
        # Email subject *must not* contain newlines
        subject = ''.join(subject.splitlines())
        body = loader.render_to_string(self.email_template_name, data)
        email_message = EmailMultiAlternatives(subject, body, self.from_email,
                                               [user.email])
        if self.html_email_template_name is not None:
            html_email = loader.render_to_string(self.html_email_template_name,
                                                 data)
            email_message.attach_alternative(html_email, 'text/html')
        email_message.send()


@rest_view("/auth/signup")
class SignupView(SendMessageTokenMixin, CreateRestViewMixin, BaseRestView):
    email_template_name = 'registration/welcome_email.txt'
    subject_template_name = 'registration/welcome_subject.txt'
    token_generator = default_token_generator
    extra_email_context = None
    from_email = None
    html_email_template_name = None
    default_public = True

    queryset = UserModel._default_manager
    fields = ['first_name', 'last_name', 'username', 'email']

    def validate(self, resolver, **kwargs):
        resolver.object.set_unusable_password()
        super().validate(resolver, **kwargs)

    def is_valid(self, resolver, created=True):
        r = super().is_valid(resolver, created)
        self.send_token_message(resolver.object)
        return r


@rest_view("/auth")
class ForgotenPasswordView(SendMessageTokenMixin, BodyMixin, BaseRestView):
    email_template_name = 'registration/password_reset_email.txt'
    subject_template_name = 'registration/password_reset_subject.txt'
    token_generator = default_token_generator
    extra_email_context = None
    from_email = None
    html_email_template_name = None

    fields = ['first_name', 'last_name', 'username', 'email']

    @method_decorator(sensitive_post_parameters())
    @method_decorator(never_cache)
    @rest_api("/forgotten-password", public=True, method="post")
    def forgotten_password(self, *args, **kwargs):
        data = self.deserialize_body()
        try:
            user = UserModel._default_manager.get(email=data['email'])
        except UserModel.DoesNotExist:
            user = None
        if user is not None:
            self.send_token_message(user)
            return JsonResponse({
                'message': _('A message has been sent')
            })
        raise ModelNotFoundAPIError(UserModel)



from datetime import date
from unittest.mock import patch

from django.test import TestCase

from pfx.pfxcore.test import APIClient, TestAssertMixin, print_response
from tests.models import Author


class TestTools(TestAssertMixin, TestCase):

    def setUp(self):
        self.client = APIClient(default_locale='en')

    @classmethod
    def setUpTestData(cls):
        cls.author1 = Author.objects.create(
            first_name='John Ronald Reuel',
            last_name='Tolkien',
            slug='jrr-tolkien')
        pass

    @patch('builtins.print')
    def test_key_error(self, mock_print):
        response = self.client.get('/api/authors/filters')
        self.assertRC(response, 200)
        with self.assertRaises(Exception):
            self.assertJE(response, 'items.@0.name-not-exists', 'not exists')

    @patch('builtins.print')
    def test_index_error(self, mock_print):
        response = self.client.get('/api/authors/filters')
        self.assertRC(response, 200)
        with self.assertRaises(Exception):
            self.assertJE(response, 'items.@99.name', 'gender')

    @patch('builtins.print')
    def test_print_response(self, mock_print):
        response = self.client.get(f'/api/authors/{self.author1.pk}')

        self.assertRC(response, 200)
        print_response(response)

        mock_print.assert_any_call(
            '\n\n*********************http response*********************')
        mock_print.assert_any_call('Status : ', 200, 'OK')
        mock_print.assert_any_call('Headers : ')
        mock_print.assert_any_call("""  Content-Type: application/json
  Content-Length: %s
  X-Content-Type-Options: nosniff
  Referrer-Policy: same-origin
  Vary: Accept-Language
  Content-Language: en
""" % (response.headers['Content-Length']))
        mock_print.assert_any_call('Content : ')
        mock_print.assert_any_call("""{
    "books": [],
    "created_at": "%s",
    "first_name": "John Ronald Reuel",
    "gender": {
        "label": "Male",
        "value": "male"
    },
    "last_name": "Tolkien",
    "meta": {},
    "name_length": 25,
    "pk": %i,
    "resource_name": "John Ronald Reuel Tolkien",
    "slug": "jrr-tolkien"
}""" % (date.today(), self.author1.pk))
        mock_print.assert_any_call(
            '*******************************************************\n\n')



from django_request_mapping import UrlPattern

from . import views

urlpatterns = UrlPattern()
urlpatterns.register(views.LocaleRestView)
urlpatterns.register(views.AuthenticationView)
urlpatterns.register(views.SignupView)
urlpatterns.register(views.ForgotenPasswordView)



from datetime import date
from functools import reduce

from django.db.models import Q
from django.utils.translation import gettext_lazy as _

from pfx.pfxcore.shortcuts import parse_bool


class FilterGroup():
    def __init__(self, name, label, filters):
        self.name = name
        self.label = label
        self.filters = filters

    @property
    def meta(self):
        return dict(name=self.name, label=self.label, items=[
            f.meta for f in self.filters
        ])

    def query(self, params):
        return Q(*[f.query(params) for f in self.filters])


class Filter():
    BooleanField = "BooleanField"
    IntegerField = "IntegerField"
    FloatField = "FloatField"
    CharField = "CharField"
    DateField = "DateField"
    ForeignKey = "ForeignKey"

    def __init__(
            self, name, label, type=None, filter_func=None, choices=None,
            related_model=None):
        self.name = name
        self.label = label
        self.type = type
        self.filter_func = filter_func
        self.choices = choices
        self.related_model = related_model

    @property
    def meta(self):
        res = dict(
            label=_(self.label),
            name=self.name,
            type=self.type)
        if self.choices:
            res['choices'] = [
                dict(label=_(v), value=k) for k, v in self.choices]
        if self.related_model:
            res['related_model'] = self.related_model
        return res

    def _parse_value(self, value):
        if self.type == Filter.BooleanField:
            return parse_bool(value)
        if self.type in (Filter.IntegerField, Filter.ForeignKey):
            if value in ['', 'null', '0']:
                return None
            return value and int(value) or None
        if self.type == Filter.FloatField:
            return value and float(value) or None
        if self.type == Filter.DateField:
            return value and date.fromisoformat(value) or None
        return value

    def query(self, params):
        values = [self._parse_value(v) for v in params.getlist(self.name)]
        if values:
            return reduce(
                lambda x, y: x | y, [self.filter_func(v) for v in values])
        return Q()


class ModelFilter(Filter):
    def __init__(
            self, model, name, label=None, type=None, filter_func=None,
            choices=None, related_model=None):
        self.model = model
        self.field = model._meta.get_field(name)
        super().__init__(
            name, label or self.field.verbose_name,
            type or self.field.get_internal_type(), filter_func,
            choices or self.field.choices, related_model or (
                self.field.remote_field and
                self.field.remote_field.model.__name__))

    @property
    def meta(self):
        res = dict(
            label=_(self.label), name=self.name, type=self.type)
        if self.choices:
            res['choices'] = [
                dict(label=_(v), value=k) for k, v in self.choices]
        if self.related_model:
            res['related_model'] = self.related_model
        return res

    def query(self, params):
        values = [self._parse_value(v) for v in params.getlist(self.name)]
        if self.filter_func and values:
            return reduce(
                lambda x, y: x | y, [self.filter_func(v) for v in values])
        elif values:
            return reduce(
                lambda x, y: x | y, [Q(**{self.name: v}) for v in values])
        return Q()




import logging
import re

from django.core.exceptions import ObjectDoesNotExist

logger = logging.getLogger(__name__)


def f(tmpl, **kwargs):
    return tmpl.format(**kwargs)


def get_object(queryset, **kwargs):
    from .exceptions import ModelNotFoundAPIError
    try:
        return queryset.get(**kwargs)
    except ObjectDoesNotExist:
        raise ModelNotFoundAPIError(queryset.model)


def get_pk(obj):
    if isinstance(obj, dict) and 'pk' in obj:
        return obj['pk']
    return obj


RE_FALSE = re.compile(r'^(0|false|)$', re.IGNORECASE)


def parse_bool(value):
    return value and not RE_FALSE.match(value)



from .basic_api_errors import BasicAPIErrorTest
from .basic_api_test import BasicAPITest
from .test_auth_api import AuthAPITest
from .test_cache import TestCache
from .test_locale_api import LocaleAPITest
from .test_perms_api import PermsAPITest



import logging

from django.http.response import JsonResponse
from django.utils.translation import gettext_lazy as _

from django_request_mapping import request_mapping

from pfx.pfxcore.exceptions import APIError

logger = logging.getLogger(__name__)


def pfx_api(public=None):
    def decorator(func):
        def wrapper(self, request, *args, **kwargs):
            self.request = request
            self.kwargs = kwargs
            try:
                self.check_perm(public, func.__name__, *args, **kwargs)
                return func(self, *args, **kwargs)
            except APIError as e:
                return e.response
            except Exception as e:
                logger.exception(e)
                return JsonResponse(dict(message=_(
                    "An internal server error occured.")), status=500)
        return wrapper
    return decorator


def rest_api(*dec_args, public=None, **dec_kwargs):
    _request_mapping = request_mapping(*dec_args, **dec_kwargs)
    _pfx_api = pfx_api(public=public)

    def decorator(func):
        return _request_mapping(_pfx_api(func))
    return decorator


rest_view = request_mapping


def rest_property(string=None, type="CharField"):
    def decorator(func):
        func.short_description = string
        func.internal_type = type
        return property(func)
    return decorator



import os

SECRET_KEY = 'fake-key'
INSTALLED_APPS = [
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.postgres',
    'pfx.pfxcore',
    'tests',
]

MIDDLEWARE = [
    'pfx.pfxcore.middleware.LocaleMiddleware',
    'django.middleware.security.SecurityMiddleware',
    'django.middleware.common.CommonMiddleware',
    'pfx.pfxcore.middleware.AuthenticationMiddleware',
    'pfx.pfxcore.middleware.CookieAuthenticationMiddleware',
]

DEFAULT_AUTO_FIELD = 'django.db.models.BigAutoField'

PFX_SECRET_KEY = "fake-secret-key"
PFX_TOKEN_VALIDITY = None

ROOT_URLCONF = 'tests.urls'
APPEND_SLASH = False

EMAIL_BACKEND = 'django.core.mail.backends.console.EmailBackend'
PFX_RESET_PASSWORD_URL = (
    'http://localhost:8000/test?token={token}&uidb64={uid}')
PFX_SITE_NAME = 'Assocman'

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
    }
]



import json
import logging
import re
from functools import reduce
from http.cookies import SimpleCookie

from django.test.client import Client as DjangoClient

logger = logging.getLogger(__name__)


def print_response(r, title="http response"):
    print(f"\n\n*********************{title}*********************")
    if not r:
        print("Response is null")
    print("Status : ", r.status_code, r.reason_phrase)
    print("Headers : ")
    print("".join(f"  {k}: {v}\n" for k, v in r.headers.items()))
    print("Content : ")
    if hasattr(r, 'json_content'):
        print(json.dumps(r.json_content, indent=4, sort_keys=True))
    elif hasattr(r, 'content') and r.content:
        print(r.content)
    elif hasattr(r, 'streaming_content'):
        print("Streaming content")
    else:
        print("Response is empty")
    print("*******************************************************\n\n")


class APIClient(DjangoClient):

    token = None
    default_locale = None
    auth_cookie = None

    def __init__(self, enforce_csrf_checks=False, raise_request_exception=True,
                 default_locale=None, with_cookie=False, **defaults):
        super().__init__(enforce_csrf_checks, raise_request_exception,
                         **defaults)
        self.default_locale = default_locale
        self.with_cookie = with_cookie

    def login(self, username, password='test',
              path='/api/auth/login', locale=None):
        self.token = None
        response = self.post(
            f"{path}{self.with_cookie and '?mode=cookie' or ''}", {
                'username': username,
                'password': password}, locale=locale)
        if response.status_code == 200:
            if self.with_cookie:
                self.auth_cookie = [
                    v for k, v in response.client.cookies.items()
                    if k == 'token'][0]
                regex = r".*token=([\w\._-]*);.*"
                self.token = re.findall(regex, str(self.auth_cookie))[0]
            else:
                self.token = response.json_content['token']
            return True
        return False

    def logout(self):
        self.token = None

    def generic(self, method, path, data='',
                content_type='application/octet-stream', secure=False,
                **extra):
        if self.token:
            if self.with_cookie:
                self.cookies = SimpleCookie(
                    {'token': self.token})
            else:
                extra.update(HTTP_AUTHORIZATION=f"Bearer {self.token}")
        return super().generic(
            method, path, data, content_type, secure, **extra)

    def update_locale(self, locale, extra):
        nextra = dict(extra)
        if locale:
            nextra.update(HTTP_X_CUSTOM_LANGUAGE=f"{locale}")
        elif self.default_locale:
            nextra.update(HTTP_X_CUSTOM_LANGUAGE=f"{self.default_locale}")
        return nextra

    @staticmethod
    def decode_response(response):
        if response.headers['Content-Type'] == 'application/json':
            response.json_content = response.json()
        return response

    def get(self, path, data=None, secure=False, locale=None, **extra):
        extra = self.update_locale(locale, extra)
        return self.decode_response(super().get(path, data, secure, **extra))

    def post(self, path, data=None, content_type='application/json',
             secure=False, locale=None, **extra):
        extra = self.update_locale(locale, extra)
        return self.decode_response(
            super().post(path, data, content_type, secure, **extra))

    def head(self, path, data=None, secure=False, locale=None, **extra):
        extra = self.update_locale(locale, extra)
        return self.decode_response(
            super().head(path, data, secure, **extra))

    def trace(self, path, secure=False, locale=None, **extra):
        extra = self.update_locale(locale, extra)
        return self.decode_response(super().trace(path, secure, **extra))

    def options(self, path, data='', content_type='application/json',
                secure=False, locale=None, **extra):
        extra = self.update_locale(locale, extra)
        return self.decode_response(
            super().options(path, data, content_type, secure, **extra))

    def put(self, path, data='', content_type='application/json',
            secure=False, locale=None, **extra):
        extra = self.update_locale(locale, extra)
        return self.decode_response(
            super().put(path, data, content_type, secure, **extra))

    def patch(self, path, data='', content_type='application/json',
              secure=False, locale=None, **extra):
        extra = self.update_locale(locale, extra)
        return self.decode_response(
            super().patch(path, data, content_type, secure, **extra))

    def delete(self, path, data='', content_type='application/json',
               secure=False, locale=None, **extra):
        extra = self.update_locale(locale, extra)
        return super().delete(path, data, content_type, secure, **extra)


class TestAssertMixin:

    # assert response status code
    def assertRC(self, response, code):
        return self.assertEqual(response.status_code, code)

    def get_val(self, response, key):
        def _p(k):
            return int(k[1:]) if k[0] == '@' else k
        try:
            return reduce(lambda c, k: c[_p(k)], key.split("."),
                          response.json_content)
        except IndexError:
            print_response(response, f'Index Error for key "{key}"')
            raise Exception(f'Index Error for key "{key}"')
        except KeyError:
            print_response(response, f'Key Error for key "{key}"')
            raise Exception(f'Key Error for key "{key}"')

    # assert JSON content at key equals value
    def assertJE(self, response, key, value):
        return self.assertEqual(self.get_val(response, key), value)

    # assert JSON content at key not equals value
    def assertNJE(self, response, key, value):
        return self.assertNotEqual(self.get_val(response, key), value)

    # assert JSON array size
    def assertSize(self, response, key, size):
        return self.assertEqual(len(self.get_val(response, key)), size)



import logging

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser
from django.http import JsonResponse
from django.utils.translation import gettext_lazy as _

import jwt
from jwt import DecodeError

from pfx.pfxcore.exceptions import APIError, AuthenticationError
from pfx.pfxcore.models import CacheableMixin

logger = logging.getLogger(__name__)


class JWTTokenDecodeMixin:

    @classmethod
    def get_cached_user(cls, pk):
        UserModel = get_user_model()
        has_cache = issubclass(UserModel, CacheableMixin)
        if (has_cache):   # pragma: no cover
            user = UserModel.cache_get(pk)
            if user:
                return user
        user = UserModel._default_manager.get(pk=pk)
        if (has_cache):   # pragma: no cover
            user.cache()
        return user

    @classmethod
    def decode_jwt(cls, token):
        try:
            opts = ({"require": ["exp"]} if settings.PFX_TOKEN_VALIDITY
                    else {})
            decoded = jwt.decode(
                token, settings.PFX_SECRET_KEY,
                options=opts,
                algorithms="HS256")
            return cls.get_cached_user(decoded['pfx_user_pk'])
        except DecodeError as e:
            logger.exception(e)
            raise AuthenticationError(message=_('Authentication error'))
        except jwt.ExpiredSignatureError:
            raise AuthenticationError(message=_('Token has expired'))
        except Exception as e:  # pragma: no cover
            logger.exception(e)
            raise AuthenticationError(message=_('Authentication error'),
                                      status=500)


class AuthenticationMiddleware(JWTTokenDecodeMixin):

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        authorization = request.headers.get('Authorization')
        if authorization:
            try:
                _, key = authorization.split("Bearer ")
            except ValueError:  # pragma: no cover
                key = None
            try:
                request.user = self.decode_jwt(key)
            except get_user_model().DoesNotExist:   # pragma: no cover
                return JsonResponse(
                    {'message': _("Authentication error")}, status=401)
            except APIError as e:
                return e.response
        else:
            if not hasattr(request, 'user'):
                request.user = AnonymousUser()

        return self.get_response(request)


class CookieAuthenticationMiddleware(JWTTokenDecodeMixin):

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        key = request.COOKIES.get('token')
        if key:
            try:
                request.user = self.decode_jwt(key)
            except get_user_model().DoesNotExist:   # pragma: no cover
                return JsonResponse(
                    {'message': 'Authentication error'}, status=401)
            except APIError as e:
                return e.response
        else:
            if not hasattr(request, 'user'):   # pragma: no cover
                request.user = AnonymousUser()
        return self.get_response(request)



from django.http import JsonResponse
from django.utils.translation import gettext_lazy as _

from .shortcuts import f


class APIError(Exception):
    def __init__(self, message, status=400, **kwargs):
        self.status = status
        self.data = dict(message=message, **kwargs)

    @property
    def response(self):
        return JsonResponse(self.data, status=self.status)


class ModelNotFoundAPIError(APIError):
    def __init__(self, model, **kwargs):
        super().__init__(
            f(_("{model} not found."), model=model._meta.verbose_name),
            status=404, **kwargs)


class AuthenticationError(APIError):
    def __init__(self, message=None, status=401, **kwargs):
        super().__init__(
            f(message or _("Authentication Error")),
            status=status, **kwargs)


class UnauthorizedError(APIError):
    def __init__(self, **kwargs):
        super().__init__(
            f(_("Unauthorized")),
            status=401, **kwargs)


class ForbiddenError(APIError):
    def __init__(self, **kwargs):
        super().__init__(
            f(_("Forbidden")),
            status=403, **kwargs)



import logging
import operator

from django.core.exceptions import ObjectDoesNotExist, ValidationError
from django.db import transaction
from django.db.models.constants import LOOKUP_SEP

logger = logging.getLogger(__name__)


class PropertyField:
    def __init__(self, method):
        self.name = method
        self.verbose_name = method.fget.short_description
        self.internal_type = method.fget.internal_type

    def get_internal_type(self):
        return self.internal_type


class ResolverValidationError(Exception):
    def __init__(self, errors):
        super().__init__("Validation errors")
        self.errors = errors


class MetaResolver:
    def __init__(self, model):
        self.model = model

    def get_field(self, lookup):
        path = lookup.split(LOOKUP_SEP)
        path, field_name = path[:-1], path[-1]
        model = self.model
        for e in path:
            model = model._meta.get_field(e).related_model
        attr = getattr(model, field_name)
        if isinstance(attr, property):
            return PropertyField(attr)
        return model._meta.get_field(field_name)


class ObjectResolver:
    def __init__(self, object):
        self.object = object
        self.related_objects = []
        self.lookups = {}

    def get_value(self, lookup):
        try:
            return operator.attrgetter(
                lookup.replace(LOOKUP_SEP, '.'))(self.object)
        except ObjectDoesNotExist:
            return

    def set_value(self, lookup, value):
        path = lookup.split(LOOKUP_SEP)
        path, fname = '__'.join(path[:-1]), path[-1]
        if path:
            fobj = self.get_value(path)
            self.related_objects.append(fobj)
        else:
            fobj = self.object
        self.lookups[fname] = lookup
        attr = getattr(type(fobj), fname)
        if isinstance(attr, property) and attr.fset is None:
            # Ignore property if there is no explicit setter.
            return
        setattr(fobj, fname, value)

    def set_values(self, **values):
        for attr, val in values.items():
            self.set_value(attr, val)

    def validate(self, **kwargs):
        errors = {}
        for obj in [*self.related_objects, self.object]:
            try:
                obj.full_clean(**kwargs)
            except ValidationError as e:
                errors.update({
                    self.lookups.get(k, k): v
                    for k, v in e.message_dict.items()})
        if errors:
            raise ResolverValidationError(errors)

    def save(self):
        if self.related_objects:
            with transaction.atomic():
                for obj in [*self.related_objects, self.object]:
                    obj.save()
        else:
            self.object.save()



import logging
import re
from datetime import datetime, timedelta
from http.cookies import SimpleCookie

from django.conf import settings
from django.contrib.auth.models import User
from django.contrib.auth.tokens import default_token_generator
from django.core import mail
from django.test import TransactionTestCase, override_settings
from django.utils.encoding import force_bytes
from django.utils.http import urlsafe_base64_encode

import jwt

from pfx.pfxcore.test import APIClient, TestAssertMixin

logger = logging.getLogger(__name__)


class AuthAPITest(TestAssertMixin, TransactionTestCase):

    def setUp(self):
        self.client = APIClient(default_locale='en')
        self.cookie_client = APIClient(default_locale='en', with_cookie=True)
        self.user1 = User.objects.create_user(
            username='jrr.tolkien',
            email="jrr.tolkien@oxford.com",
            password='RIGHT PASSWORD',
            first_name='John Ronald Reuel',
            last_name='Tolkien',
        )

    def test_invalid_login(self):
        response = self.client.post(
            '/api/auth/login', {
                'username': 'jrr.tolkien',
                'password': 'WRONG PASSWORD'})
        self.assertRC(response, 401)

    def test_valid_login(self):
        response = self.client.post(
            '/api/auth/login', {
                'username': 'jrr.tolkien',
                'password': 'RIGHT PASSWORD'})

        self.assertRC(response, 200)
        decoded = jwt.decode(
            response.json_content['token'], settings.PFX_SECRET_KEY,
            algorithms="HS256")
        self.assertEqual(decoded['pfx_user_pk'], self.user1.pk)

    def test_valid_login_with_cookie(self):
        response = self.client.post(
            '/api/auth/login?mode=cookie', {
                'username': 'jrr.tolkien',
                'password': 'RIGHT PASSWORD'})

        cookie = [v for k, v in response.client.cookies.items()
                  if k == 'token'][0]
        regex = r".*token=([\w\._-]*);.*"
        token = re.findall(regex, str(cookie))[0]

        self.assertRC(response, 200)
        decoded = jwt.decode(
            token, settings.PFX_SECRET_KEY,
            algorithms="HS256")
        self.assertEqual(decoded['pfx_user_pk'], self.user1.pk)

    def test_valid_change_password(self):
        self.client.login(
                username='jrr.tolkien',
                password='RIGHT PASSWORD')
        response = self.client.post(
            '/api/auth/change-password', {
                'old_password': 'RIGHT PASSWORD',
                'new_password': 'NEW RIGHT PASSWORD'})
        self.assertRC(response, 200)

        self.cookie_client.login(
                username='jrr.tolkien',
                password='RIGHT PASSWORD')
        response = self.client.post(
            '/api/auth/change-password', {
                'old_password': 'NEW RIGHT PASSWORD',
                'new_password': 'RIGHT PASSWORD'})
        self.assertRC(response, 200)

    def test_invalid_change_password(self):
        self.client.login(
                username='jrr.tolkien',
                password='RIGHT PASSWORD')
        response = self.client.post(
            '/api/auth/change-password', {
                'old_password': 'WRONG PASSWORD',
                'new_password': 'NEW RIGHT PASSWORD'})
        self.assertRC(response, 401)

    @override_settings(PFX_TOKEN_VALIDITY={'minutes': 30})
    def test_valid_token_with_expiration(self):
        self.client.login(
                username='jrr.tolkien',
                password='RIGHT PASSWORD')
        response = self.client.get(
            '/api/books')
        self.assertRC(response, 200)

    @override_settings(PFX_TOKEN_VALIDITY={'minutes': 30})
    def test_valid_cookie_token_with_expiration(self):
        self.cookie_client.login(
                username='jrr.tolkien',
                password='RIGHT PASSWORD')

        regex = r".*expires=([^;]*);.*"
        expires = re.findall(regex, str(self.cookie_client.auth_cookie))[0]
        d = datetime.strptime(expires, '%a, %d %b %Y %H:%M:%S %Z')

        regex = r".*Max-Age=([^;]*);.*"
        max_age = re.findall(regex, str(self.cookie_client.auth_cookie))[0]

        # cookie expires in 30 minutes +/- 5 minutes.
        self.assertTrue(
            datetime.utcnow() + timedelta(minutes=25) <
            d < datetime.utcnow() + timedelta(minutes=35))
        self.assertEqual(int(max_age), 1800)

        response = self.cookie_client.get(
            '/api/books')
        self.assertRC(response, 200)

    @override_settings(PFX_TOKEN_VALIDITY={'minutes': -30})
    def test_expired_token(self):
        self.client.login(
                username='jrr.tolkien',
                password='RIGHT PASSWORD')

        response = self.client.get(
            '/api/books')
        self.assertRC(response, 401)

    @override_settings(PFX_TOKEN_VALIDITY={'minutes': -30})
    def test_expired_cookie_token(self):
        self.cookie_client.login(
                username='jrr.tolkien',
                password='RIGHT PASSWORD')

        regex = r".*expires=([^;]*);.*"
        expires = re.findall(regex, str(self.cookie_client.auth_cookie))[0]
        d = datetime.strptime(expires, '%a, %d %b %Y %H:%M:%S %Z')

        regex = r".*Max-Age=([^;]*);.*"
        max_age = re.findall(regex, str(self.cookie_client.auth_cookie))[0]

        # cookie expires now +/- 5 minutes.
        self.assertTrue(
            datetime.utcnow() - timedelta(minutes=5) <
            d < datetime.utcnow() + timedelta(minutes=5))
        self.assertEqual(int(max_age), 0)

        response = self.cookie_client.get(
            '/api/books')
        self.assertRC(response, 401)

    def test_invalid_token(self):
        token = jwt.encode(
            {'pfx_user_pk': 1}, "A WRONG SECRET", algorithm="HS256")
        logging.disable(logging.CRITICAL)
        response = self.client.get(
            '/api/books',
            HTTP_AUTHORIZATION='Bearer ' + token,
            content_type='application/json')
        logging.disable(logging.NOTSET)
        self.assertRC(response, 401)

    def test_invalid_cookie_token(self):
        token = jwt.encode(
            {'pfx_user_pk': 1}, "A WRONG SECRET", algorithm="HS256")
        logging.disable(logging.CRITICAL)
        self.client.cookies = SimpleCookie({'token': token})
        response = self.client.get(
            '/api/books',
            content_type='application/json')
        logging.disable(logging.NOTSET)
        self.assertRC(response, 401)

    def test_signup(self):
        # Try to create with an existing username
        response = self.client.post(
            '/api/auth/signup', {
                'username': 'jrr.tolkien',
                'email': "jrr.tolkien@oxford.com",
                'first_name': 'John Ronald Reuel',
                'last_name': 'Tolkien',
            })

        self.assertRC(response, 422)
        self.assertEqual(response.json_content['username'],
                         ['A user with that username already exists.'])

        # Then create another valid user
        response = self.client.post(
            '/api/auth/signup', {
                'username': 'isaac.asimov',
                'email': 'isaac.asimov@bu.edu',
                'first_name': 'Isaac',
                'last_name': 'Asimov',
            })

        self.assertRC(response, 200)

        # Must send a welcome email
        self.assertEqual(
            mail.outbox[0].subject,
            f'Welcome on {settings.PFX_SITE_NAME}')

        self.client.logout()

        # Test that the token and uid are valid.
        regex = r"token=(.*)&uidb64=(.*)"
        token, uidb64 = re.findall(regex, mail.outbox[0].body)[0]
        response = self.client.post(
            '/api/auth/set-password', {
                'token': token,
                'uidb64': uidb64,
                'password': 'test',
                'check_password': 'test'
            })
        self.assertRC(response, 200)
        self.assertJE(response, 'message', 'password updated successfully')

    def test_forgotten_password(self):
        # Try with an nonexistent email
        response = self.client.post(
            '/api/auth/forgotten-password', {
                'email': 'isaac.asimov@bu.edu',
            })

        self.assertRC(response, 404)

        # Then try with a valid email
        response = self.client.post(
            '/api/auth/forgotten-password', {
                'email': 'jrr.tolkien@oxford.com',
            })

        self.assertRC(response, 200)

        # Must send a reset password email
        self.assertEqual(
            mail.outbox[0].subject,
            f'Password reset on {settings.PFX_SITE_NAME}')

        self.client.logout()

        # Test that the token and uid are valid.
        regex = r"token=(.*)&uidb64=(.*)"
        token, uidb64 = re.findall(regex, mail.outbox[0].body)[0]
        response = self.client.post(
            '/api/auth/set-password', {
                'token': token,
                'uidb64': uidb64,
                'password': 'test',
                'check_password': 'test'
            })
        self.assertRC(response, 200)
        self.assertJE(response, 'message', 'password updated successfully')

    def test_set_password(self):
        # Try with a wrong token and uid
        response = self.client.post(
            '/api/auth/set-password', {
                'token': 'WRONG TOKEN',
                'uidb64': 'WRONG UID',
                'password': 'NEW PASSWORD',
            })

        self.assertRC(response, 401)

        # Then try with a valid token and uid
        token = default_token_generator.make_token(self.user1)
        uid = urlsafe_base64_encode(force_bytes(self.user1.pk))
        response = self.client.post(
            '/api/auth/set-password', {
                'token': token,
                'uidb64': uid,
                'password': 'NEW PASSWORD',
            })
        self.assertRC(response, 200)



from django.urls import include, path

from django_request_mapping import UrlPattern

from . import views

handler404 = 'pfx.pfxcore.views.resource_not_found'

urlpatterns = UrlPattern()
urlpatterns.register(views.AuthorRestView)
urlpatterns.register(views.PrivateEditAuthorRestView)
urlpatterns.register(views.PrivateAuthorRestView)
urlpatterns.register(views.AdminEditAuthorRestView)
urlpatterns.register(views.AdminAuthorRestView)
urlpatterns.register(views.BookRestView)
urlpatterns.register(views.BookTypeRestView)
urlpatterns.register(views.Testi18nView)
urlpatterns.register(views.TestErrorView)

urlpatterns = [
    path('api/', include(urlpatterns)),
    path('api/', include('pfx.pfxcore.urls'))
]



import logging

from django.core.exceptions import ImproperlyConfigured
from django.db.models import Q
from django.http import JsonResponse
from django.utils.translation import ugettext as _

from pfx.pfxcore.decorator import rest_api, rest_view
from pfx.pfxcore.views import (
    BaseRestView,
    Filter,
    FilterGroup,
    ModelFilter,
    RestView,
)
from pfx.pfxcore.views.rest_views import SlugDetailRestViewMixin

from .models import Author, Book, BookType

logger = logging.getLogger(__name__)


class AuthorRestViewMixin():
    queryset = Author.objects
    fields = ['first_name', 'last_name', 'name_length',
              'books', 'created_at', 'slug', 'gender']
    list_fields = ['first_name', 'last_name', 'gender']
    readonly_fields = ['created_at']


def heroic_fantasy_filter(value):
    q = Q(last_name="Tolkien")
    return value and q or ~q


def last_name_filter(value):
    return Q(last_name__isnull=not value)


def last_name_choices_filter(value):
    return Q(last_name=value)


def pub_date_gte_filter(value):
    return Q(pub_date__gte=value)


def author_pk_filter(value):
    return Q(author__pk=value)


@rest_view("/authors")
class AuthorRestView(AuthorRestViewMixin, SlugDetailRestViewMixin, RestView):
    default_public = True
    filters = [
        FilterGroup('book_gender', _("Book Gender"), [
            ModelFilter(Author, 'science_fiction'),
            Filter(
                'heroic_fantasy', _("Heroic Fantasy"),
                Filter.BooleanField, heroic_fantasy_filter),
        ]),
        FilterGroup('custom', _("Custom"), [
            ModelFilter(
                Author, 'last_name', type=Filter.BooleanField,
                filter_func=last_name_filter),
            ModelFilter(Author, 'first_name'),
            ModelFilter(Author, 'gender'),
            Filter(
                'last_name_choices', _("Tolien or Asimov"),
                Filter.CharField, last_name_choices_filter,
                choices=[('Tolkien', "Tolkien"), ('Asimov', "Asimov")]),
        ]),
    ]

    def search_filter(self, search):
        return (
            Q(first_name__unaccent__icontains=search) |
            Q(last_name__unaccent__icontains=search))

    @rest_api("/cache/<int:id>", method="get")
    def cache_get(self, id, *args, **kwargs):
        book = Author.cache_get(id)
        if book:
            return self.response(book, from_cache=True)
        book = self.get_object(pk=id)
        book.cache()
        return self.response(book, from_cache=False)


@rest_view("/private-edit/authors")
class PrivateEditAuthorRestView(AuthorRestViewMixin, RestView):
    get_public = True
    get_list_public = True


@rest_view("/private/authors")
class PrivateAuthorRestView(AuthorRestViewMixin, RestView):
    pass


@rest_view("/admin-edit/authors")
class AdminEditAuthorRestView(AuthorRestViewMixin, RestView):
    def put_perm(self, id, *args, **kwargs):
        return self.request.user.is_superuser


@rest_view("/admin/authors")
class AdminAuthorRestView(AuthorRestViewMixin, RestView):
    def perm(self):
        return self.request.user.is_superuser


@rest_view("/books")
class BookRestView(RestView):
    queryset = Book.objects
    fields = ['name', 'author', 'pub_date', 'created_at', 'type']
    readonly_fields = ['created_at']
    default_public = True
    filters = [
        FilterGroup('custom', _("Custom"), [
            ModelFilter(Book, 'author'),
            ModelFilter(Book, 'type'),
            ModelFilter(Book, 'pages'),
            ModelFilter(Book, 'rating'),
            ModelFilter(Book, 'pub_date'),
            Filter(
                'pub_date_gte', _("Publication Date greater than"),
                Filter.DateField, pub_date_gte_filter),
            Filter(
                'author_pk', _("Author PK"),
                Filter.ForeignKey, author_pk_filter,
                related_model="Author"),
        ]),
    ]


@rest_view("/book-types")
class BookTypeRestView(RestView):
    queryset = BookType.objects
    fields = ['name', 'slug']
    readonly_fields = ['created_at']
    default_public = True


@rest_view("/test-i18n")
class Testi18nView(BaseRestView):
    default_public = True

    @rest_api("", method="get")
    def get(self, *args, **kwargs):
        return JsonResponse({'Monday': _("Monday")})


@rest_view("/error")
class TestErrorView(BaseRestView):
    default_public = True

    @rest_api("/500", method="get")
    def raise_500(self):
        raise ImproperlyConfigured("Test exception")



from django.db import models
from django.utils.functional import cached_property

from pfx.pfxcore.decorator import rest_property
from pfx.pfxcore.models import CacheableMixin, CacheDependsMixin, JSONReprMixin


class Author(CacheableMixin, models.Model):
    CACHED_PROPERTIES = ['books_count']

    first_name = models.CharField("First Name", max_length=30)
    last_name = models.CharField("Last Name", max_length=30)
    slug = models.SlugField("Slug", unique=True)
    gender = models.CharField("Gender", max_length=10, choices=[
        ('male', "Male"), ('female', "Female")], default='male')
    science_fiction = models.BooleanField("Science Fiction", default=False)
    created_at = models.DateField("Created at", auto_now_add=True)

    class Meta:
        verbose_name = "Author"
        verbose_name_plural = "Authors"

    def __str__(self):
        return f"{self.first_name} {self.last_name}"

    @rest_property("Name Length", "IntegerField")
    def name_length(self):
        return len(str(self))

    @cached_property
    def books_count(self):
        return self.books.count()


class BookType(CacheDependsMixin, JSONReprMixin, models.Model):
    CACHE_DEPENDS_FIELDS = ['books.author']

    name = models.CharField("Name", max_length=30)
    slug = models.SlugField("Slug")

    class Meta:
        verbose_name = "Book Type"
        verbose_name_plural = "Book Types"

    def __str__(self):
        return f"{self.name}"


class Book(CacheDependsMixin, JSONReprMixin, models.Model):
    CACHE_DEPENDS_FIELDS = ['author']

    name = models.CharField("Name", max_length=30)
    author = models.ForeignKey(
        'tests.Author', on_delete=models.RESTRICT,
        related_name='books', verbose_name="Author")
    type = models.ForeignKey(
        'tests.BookType', on_delete=models.SET_NULL, null=True, blank=True,
        related_name='books', verbose_name="Book Type")
    pub_date = models.DateField("Pub Date")
    created_at = models.DateField("Created at", auto_now_add=True)
    pages = models.IntegerField("Pages", null=True, blank=True)
    rating = models.FloatField("Rating", null=True, blank=True)
    reference = models.CharField(
        "Reference", max_length=30, null=True, blank=True)

    class Meta:
        verbose_name = "Book"
        verbose_name_plural = "Books"

    def __str__(self):
        return f"{self.name}"
